@@ -481,7 +481,10 @@
 
     # Creates a sustainability tracker and stores it in the private `sustainability_tracker` attribute
     create_sustain_tracker = function() {
-      if(utils::compareVersion(codecarbon["__version__"],"2.8.0")>=0){
+
+      tmp_code_carbon<-reticulate::import("codecarbon")
+      codecarbon_version=as.character(tmp_code_carbon["__version__"])
+      if(utils::compareVersion(codecarbon_version,"2.8.0")>=0){
         path_look_file=codecarbon$lock$LOCKFILE
         if(file.exists(path_look_file)){
           unlink(path_look_file)
